
from collections import defaultdict
import yaml
from explay.post_func import common_funcs


def is_buildin(func_str):
    try:
        eval(func_str)
        return True
    except NameError as ex:
        return False


def replace_str(my_string, spans, replaced_words):
    cursor = 0
    new_string = ''
    for span, word in zip(spans, replaced_words):
        prev_str = my_string[cursor:span[0]]
        next_str = my_string[span[0]:span[1]]
        new_string += prev_str + str(word)
        cursor = span[1]
    new_string += my_string[cursor:]
    return new_string


def register_custom_func(name, func):
    global common_funcs
    common_funcs[name] = func


def register_func():
    import func
    funcs = [f for f in dir(func) if f.startswith('exp')]
    for func_name in funcs:
        func_name_in_yml = func_name[4:]
        register_custom_func(func_name_in_yml, getattr(func, func_name))



def to_yml(files, set_defualtdict=False, default_type=str):
    content = yaml.load(open(files, 'r').read()) 
    if set_defualtdict:
        content = defaultdict(default_type, content)
    return content



#  def pd_set_option(pd, max_colwidth, max_columns, precision=1):
def pd_set_option(max_colwidth, max_columns, precision=1):
    import pandas as pd
    pd.set_option('display.expand_frame_repr', False)
    pd.set_option('display.max.colwidth', max_colwidth)
    pd.set_option('display.max_columns', max_columns)
    pd.set_option('precision', precision)
    pd.set_option('display.float_format', '{:20,.1f}'.format)
    pd.set_option('display.unicode.east_asian_width', True)



def get_local_variables(dir):
    print(globals().keys())
    var_excludes = ['In', 'Out', 'exit', 'quit']
    v = sorted(filter(lambda s:not s.startswith('_'), dir))
    v = list(filter(lambda x: x not in var_excludes, v))
    return v



# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import re
import sys
import json
import glob
import yaml
import regex
import __main__
from copy import copy
import datetime
from collections import namedtuple, defaultdict
import inspect
import functools
import glob
import numpy as np

import xlrd
from openpyxl import load_workbook
from openpyxl.utils.dataframe import dataframe_to_rows
from openpyxl.utils import get_column_letter
from openpyxl.styles import PatternFill, Font

from pandas import DataFrame
import pandas as pd
import pdb

from explay.utils import is_buildin, replace_str, to_yml
from explay.openpyxl_ext import insert_rows
from explay.parser import xlParser, xlBinaryParser
from explay.merger import xlMerger, xlConverter

from explay.utils import register_func
from explay.agg_func import agg_functions
from explay.post_func import common_funcs



def compose(*functions):
    def compose2(f, g):
        return lambda x: f(g(x))
    return functools.reduce(compose2, functions, lambda x: x)


class xlRenderer():
    def __init__(self, params):
        self.params = params
        self.first_row, self.idx_colname = params['first_row'], params['idx_colname']

    def __repr__(self):
        msg = '[Renderer]\n'
        msg += yaml.dump(self.params, indent=True, 
                allow_unicode=True, default_flow_style=False)
        return msg

    @classmethod
    def to_excel(cls, df, path):
        writer = pd.ExcelWriter(path, engine='xlsxwriter')
        df.to_excel(writer, sheet_name='Sheet1', index=False)

        workbook = writer.book
        worksheet = writer.sheets['Sheet1']

        datetime_format = workbook.add_format({'num_format': "yyyy-mm-dd"})
        time_format = workbook.add_format({'num_format': "hh:mm"})

        for index, values in df.iterrows():
            for i, v in enumerate(values):
                if type(v) == pd.Timestamp:
                    worksheet.write_datetime('%s%s' % (chr(65+i), index + 2), v, datetime_format)
                elif type(v) == datetime.time:
                    cell = '%s%s' % (chr(65+i), index+2)
                    worksheet.write_datetime('%s%s' % (chr(65+i), index + 2), v, time_format)
        writer.save()

    def render_excel(self, df, saved_name, template_name):

        xls_file, xlsx_file = ['template/%s.%s' % (template_name, ext) for ext in ['xls', 'xlsx']]
        template = load_workbook(xlsx_file)

        rows = dataframe_to_rows(df, index=True, header=False)
        ws = template.active
        ws.insert_rows = insert_rows

        ws.insert_rows(ws, self.first_row, len(df))
        xls_from = xlrd.open_workbook(xls_file, formatting_info=True)
        xls_sheet = xls_from.sheet_by_index(0)
        color_index = lambda row, col: (
            xls_from.xf_list[xls_sheet.cell_xf_index(row,col)].background.pattern_colour_index)
        rgb = lambda row, col: xls_from.colour_map[color_index(row,col)]
        to_hexcode = lambda rgb_code: '00%s' % ''.join([('%x'% e).upper() for e in rgb_code])

        get_cell = lambda sht, row, col: sht['%s%d' % (get_column_letter(col), row)]

        for r_idx, row in enumerate(rows, self.first_row):
            for c_idx, df_colname in self.idx_colname.items():
                col_idx = df.columns.get_loc(df_colname)
                get_cell(ws, r_idx, c_idx).value = row[col_idx+1]

        for row in ws.rows:
            max_contents_len = []
            for cell in row:
                r_idx, c_idx = cell.row, cell.col_idx

                max_content_len = len(str(get_cell(ws, r_idx, c_idx)))
                max_contents_len.append(max_content_len)

                #  if cell.has_style:
                new_cell = get_cell(ws, r_idx, c_idx)

                if r_idx >= self.first_row:
                    font = get_cell(ws, self.first_row, c_idx).font
                else:
                    font = cell.font
                new_cell.font = copy(font)
                new_cell.border = copy(cell.border)

                # uses xlrd to get cell styles
                if r_idx < self.first_row:
                    color = to_hexcode(rgb(r_idx-1, c_idx-1))
                else:
                    color = to_hexcode(rgb(self.first_row-1, c_idx-1))
                fill = PatternFill(fill_type='solid',start_color=color)
                new_cell.fill = copy(fill)

                new_cell.number_format = cell.number_format
                new_cell.protection = copy(cell.protection)
                new_cell.alignment = copy(cell.alignment)

        template.save(saved_name)
        print('{} saved'.format(saved_name))


class xlTemplate():
    def __init__(self, params):
        self.params = params
        self._template = params['template']
        self.output = params['output']

    def __getitem__(self, index):
        return self._template[index]

    def __len__(self):
        return len(self._template)


class xlManager():

    def __init__(self, yml_file, home=None, local_imported=None):
        home = os.getcwd() if not home else home
        yml_path = '%s/%s' % (home, yml_file)
        self.yml, self.home = yml_path, home
        self.sources = dict()
        self.content = _c = to_yml(yml_file, True)
        self.converter = _c['xlconverter']
        self.parser = _c['xlparser']
        self.renderer = _c['xlrenderer']
        self.renderer = xlRenderer(self.renderer) if self.renderer else None

        self.load_parser()
        if local_imported:
            self.import_local(local_imported)

    def load_parser(self):
        parsers = {}
        for each in self.parser:
            print('\n')
            name = each['name']
            each_parser = None
            each = defaultdict(str, each)
            each_parser = xlBinaryParser(each)
            parsers[name] = each_parser
        self.parsers = parsers

    def import_local(self, local):
        for name, obj in self.parsers.items():
            local[name] = obj

    def import_source_df(self, local):
        for name, obj in self.parsers.items():
            local[name] = obj

    def import_variables(self, local):
        for v, value in self.variables.items():
            local[v] = value

    def register_func(self):
        import func
        #  print('reigister_func')
        funcs = [f for f in dir(func) if f.startswith('exp')]
        for func_name in funcs:
            func_name_in_yml = func_name[4:]
            #  print('func {} registed.'.format(func_name))
            #  print('func name in yml: {}'.format(func_name_in_yml))
            register_custom_func(func_name_in_yml, getattr(func, func_name))

    def load_excel(self, converter_name, filepath, sheet_name=0):
        cv = xlConverter(self.process.converter)
        df = cv.load_excel(converter_name, filepath, sheet_name)
        return df

    def merge_from_config(self, merger_yml_file, local_imported=None):
        merger_params = to_yml(merger_yml_file, True)['xlmerger']
        merged_all = []
        for each in merger_params:
            each = defaultdict(str, each)
            merge_type, name, location = each['type'], each['name'], each['location']
            print('\n', merge_type, name, location)
            converter_name = each['converter_name']
            if merge_type == 'merge_sheets':
                df_merged = self.merge_sheets(converter_name, location, each['sheet_names'])
            elif merge_type == 'merge_all':
                df_merged, _ = self.merge_all(converter_name, each['sheet_name'])
            elif merge_type == 'merge_files':
                df_merged = self.merge_files(converter_name, location, each['sheet_name'])
            merged_all.append(df_merged)
            if local_imported:
                local_imported[name] = df_merged
        return merged_all

    def merge_sheets(self, converter_name, filepath, sheet_names):
        source_path = os.path.join(self.home, 'source')
        self.merger = xlMerger(self.converter, source_path)
        df_merged = self.merger.merge_sheets(converter_name, filepath, sheet_names)
        self.sources[converter_name] = df_merged
        return df_merged

    def merge_files(self, converter_name, filepaths, sheet_name=0):
        source_path = os.path.join(self.home, 'source')
        self.merger = xlMerger(self.converter, source_path)
        df_merged = self.merger.merge_files(converter_name, filepaths, sheet_name)
        self.sources[converter_name] = df_merged
        return df_merged

    def merge_all(self, converter_name, sheet_name=0, filename_excludes=None, save=True):
        print(sheet_name, filename_excludes)
        source_path = os.path.join(self.home, 'source')
        self.merger = xlMerger(self.converter, source_path)
        df_merged, file_names = self.merger.merge_all(converter_name, sheet_name, filename_excludes)
        self.sources[converter_name] = df_merged
        print('files merged: %s' % ','.join(file_names))
        if save:
            saved_path = '{}/merged.xlsx'.format(self.home)
            self.renderer.to_excel(df_merged, saved_path)
        return df_merged, file_names

    def to_excel(self, saved_path):
        if self.df is not None:
            self.renderer.to_excel(self.df, saved_path)

    def render_excel(self, df, excel_template, saved_path):
        self.renderer.render_excel(df, saved_path, excel_template)


class ExPlay():
    def __init__(self, home=None):
        self.home = home if home else os.getcwd()
        self._sources = dict()
        self._parse_yml()
        from explay.utils import pd_set_option; pd_set_option(max_colwidth=40, max_columns=15)

    def _parse_yml(self):
        yml_files = glob.glob(f'{self.home}/*.yml')
        convs, mergs, parss, rends, projs, outs = [], [], [], [], [], []
        for f in yml_files:

            each = to_yml(f, True)
            converter = each['xlconverter']
            merger = each['xlmerger']
            parser = each['xlparser']
            renderer = each['xlrenderer']
            project = each['xlproject']
            out = each['xloutput']

            if converter: convs.append([f, converter])
            if merger: mergs.append([f, merger])
            if parser: parss.append([f, parser])
            if renderer: rends.append([f, renderer])
            if project: projs.append([f, project])
            if out: outs.append([f, out])

        assert all([len(e)<2 for e in [convs, mergs, parss, rends, projs, outs]])
        self._conv_file = convs[0][0] if convs else None
        self._merg_file = mergs[0][0] if mergs else None
        self._pars_file = parss[0][0] if parss else None
        self._rend_file = rends[0][0] if rends else None
        self._proj_file = projs[0][0] if projs else None
        self._out_file = outs[0][0] if outs else None
        self._conv_params = convs[0][1] if convs else None
        self._merg_params = mergs[0][1] if mergs else None
        self._pars_params = parss[0][1] if parss else None
        self._rend_params = rends[0][1] if rends else None
        self._proj_params = projs[0][1] if projs else None
        self._out_params = outs[0][1] if outs else None

        self._converter = xlConverter(self._conv_params) if convs else None
        self._parsers = [xlBinaryParser(defaultdict(str, each_params)) 
                for each_params in self._pars_params] if parss else []
        self._renderer = xlRenderer(self._rend_params) if rends else None

        self._template = xlTemplate(self._out_params) if outs else None
        if self._proj_params:
            self._project = yaml.dump(self._proj_params, indent=True, default_flow_style=False)
        else:
            self._project = None

    def show_config(self):
        if self._converter:
            print('************************')
            print('*       converter      *')
            print('************************')
            print(self._converter)

        if self._merg_params:
            print('************************')
            print('*         merger       *')
            print('************************')
            merg_print = yaml.dump(self._merg_params, allow_unicode=True, indent=True)
            print(merg_print)

        if self._parsers:
            print('************************')
            print('*         parser       *')
            print('************************')
            for each_parser in self._parsers:
                print(each_parser, '\n')

        if self._renderer:
            print('************************')
            print('*       renderer       *')
            print('************************')
            print(self._renderer)
        
        if self._project:
            print('************************')
            print('*       project        *')
            print('************************')
            print(self._project)
        
    def _merge_sheets(self, conv_name, xlsx_path, sheet_names):
        xlsx_dir = os.path.abspath(os.path.dirname(xlsx_path))
        self.merger = xlMerger(self._conv_params, xlsx_dir)
        df_merged = self.merger.merge_sheets(conv_name, xlsx_path, sheet_names)
        self._sources[conv_name] = df_merged
        return df_merged

    def merge_sheets(self, conv_name, xlsx_path, sheet_names, save=False):
        print('sheets of file %s merged.' % xlsx_path)
        df_merged = self._merge_sheets(conv_name, xlsx_path, sheet_names)
        if save:
            saved_path = '{}/{}_merged.xlsx'.format(self.home, conv_name)
            xlRenderer.to_excel(df_merged, saved_path)
        else:
            print(df_merged)

    def _get_abs_source_path(self, xlsx_dir=None):
        if xlsx_dir:
            if os.path.isabs(xlsx_dir):
                source_path = xlsx_dir
            else:
                source_path = os.path.join(self.home, xlsx_dir)
        else:
            source_path = self.home
        return source_path

    def _merge_files(self, conv_name, relative_paths, xlsx_dir=None, sheet_name=0):
        source_path = self._get_abs_source_path(xlsx_dir)
        filepaths = [os.path.join(source_path, f) for f in relative_paths]
        self.merger = xlMerger(self._conv_params, source_path)
        df_merged = self.merger.merge_files(conv_name, filepaths, sheet_name)
        self._sources[conv_name] = df_merged
        return df_merged

    def merge_files(self, conv_name, relative_paths, xlsx_dir=None, sheet_name=0, save=False):
        source_path = self._get_abs_source_path(xlsx_dir)
        filepaths = [os.path.join(source_path, f) for f in relative_paths]
        df_merged = self._merge_files(conv_name, relative_paths, xlsx_dir, sheet_name)
        print('files merged:\n%s' % '\n'.join(filepaths))
        if save:
            saved_path = '{}/{}_merged.xlsx'.format(self.home, conv_name)
            xlRenderer.to_excel(df_merged, saved_path)
        else:
            print(df_merged)

    def _merge_all(self, conv_name, xlsx_dir=None, sheet_name=0, excludes=None):
        source_path = self._get_abs_source_path(xlsx_dir)
        self.merger = xlMerger(self._conv_params, source_path)
        df_merged, file_names = self.merger.merge_all(conv_name, sheet_name, excludes)
        self._sources[conv_name] = df_merged
        return df_merged, file_names

    def merge_all(self, conv_name, xlsx_dir=None, sheet_name=0, excludes=None, save=False):
        df_merged, file_names = self._merge_all(conv_name, xlsx_dir, sheet_name, excludes)
        print('files merged: %s' % ','.join(file_names))
        if save:
            saved_path = '{}/{}_merged.xlsx'.format(self.home, conv_name)
            xlRenderer.to_excel(df_merged, saved_path)
        else:
            print(df_merged)

    def _df_inputs(self):
        if not self._merg_params: return None
        merged_all = {}
        for each in self._merg_params:
            each = defaultdict(str, each)
            merge_type, name, location = each['type'], each['name'], each['location']
            xlsx_dir, xlsx_path = each['xlsx_dir'], each['xlsx_path']
            print('\n', merge_type, name, location)
            converter_name = each['converter_name']

            print(merge_type)
            if merge_type == 'merge_sheets':
                df_merged = self._merge_sheets(converter_name, xlsx_path, each['sheet_names'])

            elif merge_type == 'merge_all':
                excludes = each['excludes']
                df_merged, _ = self._merge_all(converter_name, xlsx_dir,
                        each['sheet_name'], excludes)

            elif merge_type == 'merge_files':
                df_merged = self._merge_files(converter_name, location,
                        xlsx_dir, each['sheet_name'])
            merged_all[name] = df_merged
        return merged_all

    def _run(self, node):
        local_name = dir(__main__)
        _local = lambda x: getattr(__main__, x)
        if 'name' in node:
            del node['name']
        node_name = list(node.keys())[0]
        parser = getattr(__main__, node_name)
        tp = parser.check_ParserType()
        if tp=='binary_parser':
            node_child = node[node_name]
            left = node_child['left']
            right = node_child['right']
            if type(left)==str and left in local_name:
                left_result = _local(left)
            else:
                left_result = self._run(left)

            if type(right)==str and right in local_name:
                right_result = _local(right)
            else:
                right_result = self._run(right)
            return parser(left_result, right_result)

        elif tp=='unary_parser':
            node_key = list(node.keys())[0]
            node_value = node[node_key]
            if type(node_value)==str and node_value in local_name:
                result = _local(node_value)
            else:
                result = self._run(node_value)
            return parser(result)

    def run_proj(self, to_excel=True):
        components = [self._converter, self._merg_params, self._parsers, self._project]
        if not all(components):
            print('please define all explay components!')
            return

        self.export()
        self.results = {}
        for each_proj in self._proj_params:
            proj_name = each_proj['name']
            self.results[proj_name] = self._run(each_proj)

        if to_excel:
            if self._out_params and self._renderer and self._template:
                self._render_excel()
            else:
                self._to_excel()

    def _to_excel(self):
        for proj_name, each_result in self.results.items():
            self._renderer.to_excel(each_result, 
                    'out_{}.xlsx'.format(proj_name))

    def _render_excel(self):
        for e in self._template.output:
            template_name = e['template']
            proj_result = self.results[e['proj_result']]
            path=  e['path']
            self._renderer.render_excel(proj_result, path, template_name)

    def export(self):
        inputs = self._df_inputs()
        for input_name, each_df in inputs.items():
            setattr(__main__, input_name, each_df)
        for each_parser in self._parsers:
            setattr(__main__, each_parser.name, each_parser)


if __name__ == "__main__":
    
    ee = ExPlay()
    ee.export(locals())
    #  conv, pars, inputs = ee.run()

    #  for name, each_df in inputs.items():
        #  locals()[name] = each_df

    x1 = ACTION1(df, df_gender)
    x2 = ACTION2(x1, df_code)
    DF = ACTION3(ACTION1._output.output[0], x2)

    #  x1 = pars[2](df, df_gender)
    #  x2 = pars[1](x1, df_code)
    #  DF = pars[0](pars[2]._output.output[0], x2)




#  custom_funcs = {}


common_funcs = {

    # date  as datetime
    'format_date': lambda date: '%s年%s月%s日' % (date.year - 1911, date.month, date.day),
    'y_m_d': lambda date: [date.year, date.month, date.day],
    'date_year': lambda date : date.year, 
    'date_month': lambda date: date.month, 
    'date_day': lambda date: date.day, 
    
    # year as integer
    'year_minguo_to_bc': lambda year: year + 1911,
    'year_bc_to_minguo': lambda year: year - 1911
}




import json


agg_functions = {
    #  'concat': lambda x: json.dumps(format_date_series(sorted(x))),
    'list': list,
    'concat2': lambda x: json.dumps([e for e in x]),
    'join': lambda x, delimiter: delimiter.join(x),
}



# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import re
import sys
import json
import yaml
import regex
from copy import copy
import datetime
from collections import namedtuple, defaultdict
import inspect
import functools
import glob
import numpy as np

import xlrd
from openpyxl import load_workbook
from openpyxl.utils.dataframe import dataframe_to_rows
from openpyxl.utils import get_column_letter
from openpyxl.styles import PatternFill, Font

from pandas import DataFrame
import pandas as pd
import pdb

from explay.utils import is_buildin, replace_str, register_func
from explay.openpyxl_ext import insert_rows

from explay.agg_func import agg_functions
from explay.post_func import common_funcs



def rpartial(func, *args):
    return lambda *a: func(*(a + args))


def compose(*functions):
    def compose2(f, g):
        return lambda x: f(g(x))
    return functools.reduce(compose2, functions, lambda x: x)


def register_custom_func(name, func):
    global common_funcs
    common_funcs[name] = func


class Operation():
    def __init__(self, params):
        self.params = params
        self.type = params['type']
        self.args = params['args']
        self.name = params['name'] if 'name' in params else ''
        self.load()

    def load(self):
        pass
    
    def __repr__(self):
        operation_type = self.__class__.__name__
        msg = '[%s][%s]' % (operation_type, self.name)
        return msg

        #  return '<{}>'.format(self.__class__.__name__)


class UnaryOperation(Operation):
    def __init__(self, params):
        Operation.__init__(self, params)
        self.op_type = 'unary'


class BinaryOperation(Operation):
    def __init__(self, params):
        Operation.__init__(self, params)
        self.op_type = 'binary'


class Sort(Operation):
    def __init__(self, params):
        UnaryOperation.__init__(self, params)

    def load(self):
        #  print('Sort load')
        self.values = self.args['values']

    def parse(self, df):
        sort_values = self.values
        output = df.sort_values(sort_values)
        return output


class GroupBy(UnaryOperation):
    def __init__(self, params):
        UnaryOperation.__init__(self, params)

    def load(self):
        #  print('Groupby load')
        self.by = self.args['by']
        self.agg = self.args['agg']
        self.eval_func()

    def eval_func(self):
        for name, func_with_arg in self.agg.items():
            title, func_str = func_with_arg[:2]
            if len(func_with_arg) > 1:
                func_args = func_with_arg[2:]

            if not is_buildin(func_str):
                try:
                    agg_func = agg_functions[func_str]
                    if func_args:
                        agg_func = rpartial(agg_func, *func_args)
                except NameError as ex:
                    raise
            else:
                agg_func = eval(func_str)
            self.agg[name] = [title, agg_func]

    def parse(self, df):
        group_by = self.by
        agg_dict = self.agg
        D = df.groupby(group_by, sort=False)

        dataframes = []
        for name, values in agg_dict.items():
            col, func = values
            df_grouped = pd.DataFrame(D[col].apply(lambda x: func(x)))
            df_grouped.columns = [name]
            dataframes.append(df_grouped)
        output = pd.concat(dataframes, axis=1)
        output = output.reset_index(group_by)

        return output


#  class Trim(UnaryOperation):
    #  def __init__(self, params):
        #  UnaryOperation.__init__(self, params)

    #  def load(self):
        #  print('Trim load')
        #  args = self.params['args']
        #  self.columns = args['columns']
        #  self.reset_index = args['reset_index']


    #  def parse(self, df):
        #  output = df.reset_index(df.index.name)[self.columns]
        #  output.index = range(1, len(output)+1)
        #  return output


class Extension(UnaryOperation):
    def __init__(self, params):
        UnaryOperation.__init__(self, params)

    def load(self):
        #  print('Extension load')
        args = self.params['args']
        self.title = args['title']
        self.output_type = args['type']
        self.func = args['func']

    def parse(self, df):
        func_name = self.func

        if self.title in df.columns:
            df_output = df.drop(self.title, axis=1)
        else:
            df_output = df
        output = []

        for index, row in df.iterrows():
            try:
                titles = df.columns.tolist()
                input_dict = dict(zip(titles, row))

                if func_name.startswith('template@'):
                    
                    input_dict.keys()
                    template_string = func_name[9:]
                    p = regex.compile('{.*?\L<options>.*?}', options=input_dict.keys())
                    grouped = [m for m in p.finditer(template_string)]
                    spans = [list(g.span()) for g in grouped]

                    if not grouped:
                        each_output = template_string
                    else:
                        for key, value in input_dict.items():
                            locals()[key] = value

                        for key, value in common_funcs.items():
                            locals()[key] = value

                        values = []
                        for group in grouped:
                            span = group.span()
                            matched = group.group()
                            #  print('matched: %s' % matched)
                            value = eval(matched[1:-1])
                            values.append(value)
                        each_output = replace_str(template_string, spans, values)

                        cast_datetime = rpartial(datetime.datetime.strptime, '%Y-%m-%d %H:%M:%S')

                        if self.output_type:
                            cast = {'int': compose(int, float),
                                    'float': float,
                                    'list': json.loads,
                                    'str': str,
                                    'datetime': cast_datetime}
                            if self.output_type== 'list':
                                each_output = each_output.replace("'", '"')
                            each_output = cast[self.output_type](each_output)
                else:
                    func = global_func[func_name]
                    sig = inspect.signature(func)
                    arg_names = list(sig.parameters)
                    if len(arg_names)==1:
                        inputs = {arg_names[0]: input_dict[input[0]]}
                    else:
                        inputs = {k:v for k,v in input_dict.items() if k in arg_names}
                    each_output = func(**inputs)

                output.append(each_output)

            except AssertionError as ex:
                print('template後的參數未定義')
                import pdb; pdb.set_trace()
                #  raise AssertionError

        df_ext = pd.DataFrame({self.title: output}, index = df.index)
        df_output = pd.concat((df_output, df_ext), axis=1)
        return df_output


class Join(BinaryOperation):
    def __init__(self, params):
        BinaryOperation.__init__(self, params)

    def load(self):
        #  print('Join load')
        args = self.params['args']
        self.on = args['join_on']
        self.how = args['how']
        self.right_cols = args['right_cols']
        self.left_fillna = args['left_fillna']

    def parse(self, left, right):
        if self.right_cols:
            right = right[self.on + self.right_cols]
        joined_df = pd.merge(left, right, on=self.on, how=self.how)

        if self.left_fillna:
            col_to_fillna, replaced = self.left_fillna
            isnull = getattr(joined_df, col_to_fillna).isnull()
            joined_df.loc[isnull, col_to_fillna] = [ [[]] * isnull.sum() ]
        return joined_df


class Condition():
    def __init__(self, variable, operand, value):
        self.var = variable
        self.op = operand
        self.value = value

    def query(self, df):
        if self.op in ['==', 'eq']:
            new_df = df[df.apply(lambda x: x[self.var] == self.value, axis=1)]
        elif self.op == 'in':
            new_df = df[df.apply(lambda x: x[self.var] in self.value, axis=1)]
        elif self.op == 'not_in':
            new_df = df[df.apply(lambda x: x[self.var] not in self.value, axis=1)]

        return new_df


class Filter(UnaryOperation):
    def __init__(self, params):
        UnaryOperation.__init__(self, params)

    def load(self):
        #  print('Filter load')
        args = self.params['args']
        self.title = args['title']
        self.condition = args['condition']

    def parse(self, df):
        if self.condition:
            cond_type = self.condition['type']
            items = self.condition['items']
            conditions = [Condition(*item) for item in items]
            if cond_type == 'once':
                one_cond = conditions[0]
                result = one_cond.query(df)

            elif cond_type == 'and':
                pass

            elif cond_type == 'or':
                pass
        else:
            result = df

        if self.title:
            result = result[self.title]

        return result


class Pivot(UnaryOperation):
    def __init__(self, params):
        UnaryOperation.__init__(self, params)

    def load(self):
        #  print('Pivot load')
        args = self.params['args']
        self.index = args['index']
        self.columns = args['columns']
        self.values = args['values']
        self.reset_index = args['reset_index']

    def parse(self, df):
        index = self.index
        result = df.pivot(index=self.index, columns=self.columns, values=self.values)
        if self.reset_index:
            result = result.reset_index()
        return result


class Melt(UnaryOperation):
    def __init__(self, params):
        UnaryOperation.__init__(self, params)

    def load(self):
        #  print('Melt load')
        args = self.params['args']
        self.id_vars = args['id_vars'] if args['id_vars'] else None
        self.value_vars = args['value_vars'] if args['value_vars'] else None
        self.value_name = args['value_name'] if args['value_name'] else None
        self.var_name = args['var_name'] if args['var_name'] else 'value'
        
    def parse(self, df):
        if 'by_range' in self.value_vars:
            rng = self.value_vars['by_range']
            value_vars = df.columns[rng[0]-1:rng[1]-1]
        elif 'by_name' in self.value_vars:
            pass
        
        result = df.melt(id_vars=self.id_vars,
                 value_vars=value_vars,
                 var_name=self.var_name,
                 value_name=self.value_name)
        return result


def melt_parser(op_params):
    item = Melt(op_params)
    return [item]
        

def pivot_parser(op_params):
    item = Pivot(op_params)
    return [item]
        

def sort_parser(op_params):
    item = Sort(op_params)
    return [item]
        

def filter_parser(op_params):
    item = Filter(op_params)
    return [item]
        

def extend_parser(op_params):
    operations = []
    op_args = op_params['args']
    arg_title = op_args['title']
    arg_title = [arg_title] if type(arg_title)==str else arg_title
    arg_funcs = [e.rstrip() for e in op_args['func'].split('@')]
    len_title, len_func = len(arg_title), len(arg_funcs)-1

    if 'type' not in op_args:
        arg_types = [''] * len_title
    else:
        arg_types = op_args['type']
        arg_types = [arg_types] if type(arg_types)==str else arg_types

    len_type = len(arg_types)

    if len_title != len_func:
        raise Exception('ProcessError',
            'number of extension functions must be pairsed with titles!')

    elif len_title != len_type:
        raise Exception('ProcessError',
            'number of extension types must be pairsed with titles!')

    elif arg_funcs[0] != 'template':
        raise Exception('ProcessError', 'Extension Format not matched!')

    each_op_args = defaultdict(str)

    each_params = op_params.copy()
    for arg_title, arg_func, arg_type in zip(arg_title, arg_funcs[1:], arg_types):
        each_op_args['title'] = arg_title
        each_op_args['func'] = 'template@{}'.format(arg_func)
        each_op_args['type'] = arg_type

        each_params['args'] = each_op_args
        item = Extension(each_params)
        operations.append(item)

    return operations
        

def groupby_parser(op_params):
    item = GroupBy(op_params)
    return [item]


def trim_parser(op_params):
    item = Trim(op_params)
    return [item]


def join_parser(op_params):
    item = Join(op_params)
    return [item]


class xlParser():

    def __init__(self, initializer):
        self._operations = []
        self.output = []
        self._operations = self.parse_schema(initializer)
        #  register_func()

    @classmethod
    def parse_schema(cls, schema):
        _operations = []
        for each_operation in schema:
            op_type = each_operation['type']
            op_args = defaultdict(str, each_operation['args'])
            parsers = {'group_by': groupby_parser,
                       'extend': extend_parser,
                       'trim': trim_parser,
                       'join': join_parser,
                       'filter': filter_parser,
                       'melt': melt_parser,
                       'pivot': pivot_parser,
                       'sort': sort_parser,
                      }
            op_parser = parsers[op_type]
            op_params = each_operation.copy()
            op_params['args'] = op_args
            each_operations = op_parser(op_params) # could have more than one Operation
            _operations.extend(each_operations)
        return _operations

    def __getitem__(self, position):
        return self._operations[position]

    def __len__(self):
        return len(self._operations)

    def __repr__(self):
        return '\n'.join([str(e) for e in self._operations])
        
    def __call__(self, left, right=None):
        return self.parse(left, right)

    def parse(self, left, right=None):
        register_func()
        df = left
        for i, each_op in enumerate(self._operations):
            print(i, each_op)
            if each_op.op_type == 'unary':
                df = each_op.parse(df)
            else:
                df = each_op.parse(df, right)
            self.output.append(df)
        result = self.output[-1]
        return result

    def show_outputs(self, num_for_each=5):
        for i, (each, op) in enumerate(zip(self.output, self._operations)):
            print('[index: %d][%s][%s]' % (i, op, op.name))
            print(each.head(num_for_each), '\n\n')


class xlBinaryParser():

    def __init__(self, binary_initializer):
        self.name = binary_initializer['name']
        left = binary_initializer['left']
        right = binary_initializer['right']
        output = binary_initializer['output']
        self._left = xlParser(left)
        self._right = xlParser(right)
        self._output = xlParser(output)

    def num_of_binary_parser(self, operations):
        op_basetype = lambda op: op.__class__.__bases__[0]
        op_basetypes = [op_basetype(o) for o in operations]
        num = len(list(filter(lambda o: o==BinaryOperation, op_basetypes)))
        return num

    def check_ParserType(self):
        baseclass = self._output[0].__class__.__bases__[0]
        check1 = self._output is not None
        unary_check1 = (not self._left and not self._right)
        unary_check2 = self.num_of_binary_parser(self._output)==0
        binary_check1 = self.num_of_binary_parser(self._output)==1
        binary_check2 = (self.num_of_binary_parser(self._left) + 
                         self.num_of_binary_parser(self._right))==0
        if not check1:
            return None
        elif unary_check1 and unary_check2:
            return 'unary_parser'
        elif binary_check1 and binary_check2:
            return 'binary_parser'

    def __repr__(self):
        left_ops = '\n'.join([' %s' % e for e in self._left])
        right_ops = '\n'.join([' %s' % e for e in self._right])
        output_ops = '\n'.join([' %s' % e for e in self._output])
        left_repr = '[left]\n%s' % left_ops if left_ops else ''
        right_repr = '[right]\n%s' % right_ops if right_ops else ''
        output_repr = '[output]\n%s' % output_ops
        all_repr = [left_repr, right_repr, output_repr]
        all_repr = list(filter(lambda x: x, all_repr))
        all_parsers = '\n'.join(all_repr)

        if self.check_ParserType() == 'unary_parser':
            my_type = '[Unary Parser][%s]' % self.name
        elif self.check_ParserType() == 'binary_parser':
            my_type = '[Binary Parser][%s]' % self.name
        else:
            raise
        return '%s\n%s' % (my_type, all_parsers)

    def __call__(self, left, right=None):
        return self.parse(left, right)

    def parse(self, left, right=None):
        if not self._left:
            print('no left!')
        if not self._right:
            print('no right!')

        if self.check_ParserType() == 'unary_parser':
            output = self._output(left)
        else:
            left_output = self._left.parse(left) if self._left else left
            right_output = self._right.parse(right) if self._right else right
            output = self._output.parse(left_output, right_output)

        return output

    def show_outputs(self, num_for_each=5):
        print('[LEFT]')
        self._left.show_outputs(num_for_each)
        print('[RIGHT]')
        self._right.show_outputs(num_for_each)
        print('[OUTPUT]')
        self._output.show_outputs(num_for_each)



# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import re
import sys
import json
import yaml
import regex
from copy import copy
import datetime
from collections import namedtuple, defaultdict
import inspect
import functools
import glob
import numpy as np

import xlrd
from openpyxl import load_workbook
from openpyxl.utils.dataframe import dataframe_to_rows
from openpyxl.utils import get_column_letter
from openpyxl.styles import PatternFill, Font

from pandas import DataFrame
import pandas as pd
import pdb

from explay.utils import is_buildin, replace_str
from explay.openpyxl_ext import insert_rows

from explay.utils import register_func
from explay.agg_func import agg_functions
from explay.post_func import common_funcs



def compose(*functions):
    def compose2(f, g):
        return lambda x: f(g(x))
    return functools.reduce(compose2, functions, lambda x: x)


class xlConverter():
    def __init__(self, params):
        self.params = dict([(e['name'], {k:v for k,v in e.items() if k!='name'}) for e in params])

    def __repr__(self):
        msg = '[Converter]\n'
        for name, param in self.params.items():
            msg += ' converter_name: %s\n' % name
            #  for each in param:
                #  msg += str(each)
        return msg

    def _load_excel(self, filepath, sheet_name, first_row, idx_colname, resetindex=True):
        col_indexes, col_names = list(zip(*list(idx_colname.items())))

        # handle column data type
        types = []
        col_names = list(col_names)
        for i, c in enumerate(col_names):
            name_type = c.split('=')
            if len(name_type) > 1: # has type definition
                types.append(name_type[-1].strip())
                col_names[i] = name_type[0].strip()
            else:
                types.append(None)

        parse_cols = [c-1 for c in col_indexes]
        df = pd.read_excel(filepath, sheet_name=sheet_name,
                skiprows=first_row-1, header=None, usecols=parse_cols, names=col_names)
        df = df.reset_index(drop=True) if resetindex else df

        return df, types

    def load_excel(self, converter_name, filepath, sheet_name=0, resetindex=True):
        if type(converter_name)==list:
            output = []
            for each in converter_name:
                first_row = self.params[each]['first_row']
                idx_colname = self.params[each]['idx_colname']
                df, types = self._load_excel(filepath, sheet_name, first_row, idx_colname, resetindex)
                import pdb; pdb.set_trace()
                output.append(df)
        else:
            first_row = self.params[converter_name]['first_row']
            idx_colname = self.params[converter_name]['idx_colname']
            df, types = self._load_excel(filepath, sheet_name, first_row, idx_colname, resetindex)
            output = df

        if 'dropna' in self.params[converter_name]:
            df.dropna(subset=[self.params[converter_name]['dropna']], inplace=True)

        if 'trim' in self.params[converter_name]:
            cols_trim = self.params[converter_name]['trim']
            if type(cols_trim) != list:
                cols_trim = [cols_trim]
            df[cols_trim] = df[cols_trim].applymap(lambda x: str(x).strip())

        for col, col_type in zip(df.columns, types):
            if col_type:
                cast = {'int': compose(int, float), 'float': float}
                df[col] = df[col].apply(cast[col_type])

        return output


class xlMerger(xlConverter):
    def __init__(self, params, source_path):
        xlConverter.__init__(self, params)
        self.source_path = source_path

    def merge_files(self, converter_name, filepaths, sheet_name=0):
        df_all = []
        for filepath in filepaths:
            df_all.append(self.load_excel(converter_name, filepath, sheet_name))
        df = pd.concat(df_all, axis=0)
        df.index = range(len(df))
        return df

    def merge_sheets(self, converter_name, filepath, sheet_names):
        df_list = []
        for sheet_name in sheet_names:
            df_each = self.load_excel(converter_name, filepath, sheet_name)
            df_list.append(df_each)
        df_merged = pd.concat(df_list)
        return df_merged

    def merge_all(self, converter_name, sheet_name=0, filename_excludes=None):
        files = glob.glob('{}/*xlsx*'.format(self.source_path))
        file_names = [os.path.basename(f) for f in files]
        if filename_excludes:
            file_names = list(filter(lambda f: f not in filename_excludes, file_names))
        else:
            file_names = file_names

        df_all = []
        for f in file_names:
            file_path = os.path.join(self.source_path, f)
            df_all.append(self.load_excel(converter_name, file_path, sheet_name))
        df = pd.concat(df_all, axis=0)
        df.index = range(len(df))
        return df, file_names


#  class xlParser():
    #  def __init__(self, jobs):
        #  self.jobs = jobs
        #  self.outputs = defaultdict(list)

    #  def parse(self, df_input, job_name=None):
        #  if job_name in self.outputs:
            #  self.outputs[job_name] = list()

        #  for jobname, operations in self.jobs.items():
            #  if jobname != job_name: continue
            #  df = df_input

            #  for i, each_op in enumerate(operations):
                #  df = each_op.parse(df)
                #  self.outputs[jobname].append(df)

    #  def show_process(self, job_name):
        #  for i, o in enumerate(self.outputs[job_name], 1):
            #  print('\n\noperation %d' % i)
            #  print(o.head())






from openpyxl.worksheet.worksheet import Worksheet
from openpyxl.utils import get_column_letter
from openpyxl.cell import Cell
import copy
import re


def insert_rows(self, row_idx, cnt, above=False, copy_style=True):
    get_cell = lambda sht, row, col: sht['%s%d' % (get_column_letter(col), row)]

    """Inserts new (empty) rows into worksheet at specified row index.

    :param row_idx: Row index specifying where to insert new rows.
    :param cnt: Number of rows to insert.
    :param above: Set True to insert rows above specified row index.
    :param copy_style: Set True if new rows should copy style of immediately above row.
    :param fill_formulae: Set True if new rows should take on formula from immediately above row, filled with references new to rows.

    Usage:

    * insert_rows(2, 10, above=True, copy_style=False)

    """
    CELL_RE  = re.compile("(?P<col>\$?[A-Z]+)(?P<row>\$?\d+)")

    row_idx = row_idx - 1 if above else row_idx

    def replace(m):
        row = m.group('row')
        prefix = "$" if row.find("$") != -1 else ""
        row = int(row.replace("$",""))
        row += cnt if row > row_idx else 0
        return m.group('col') + prefix + str(row)

    # First, we shift all cells down cnt rows...
    old_cells = set()
    old_fas   = set()
    new_cells = dict()
    new_fas   = dict()
    for c in self._cells.values():

        old_coor = c.coordinate

        # Shift all references to anything below row_idx
        if c.data_type == Cell.TYPE_FORMULA:
            c.value = CELL_RE.sub(
                replace,
                c.value
            )
            # Here, we need to properly update the formula references to reflect new row indices
            if old_coor in self.formula_attributes and 'ref' in self.formula_attributes[old_coor]:
                self.formula_attributes[old_coor]['ref'] = CELL_RE.sub(
                    replace,
                    self.formula_attributes[old_coor]['ref']
                )

        # Do the magic to set up our actual shift
        if c.row > row_idx:
            old_coor = c.coordinate
            old_cells.add((c.row,c.col_idx))
            c.row += cnt
            new_cells[(c.row,c.col_idx)] = c
            if old_coor in self.formula_attributes:
                old_fas.add(old_coor)
                fa = self.formula_attributes[old_coor].copy()
                new_fas[c.coordinate] = fa

    for coor in old_cells:
        del self._cells[coor]
    self._cells.update(new_cells)

    for fa in old_fas:
        del self.formula_attributes[fa]
    self.formula_attributes.update(new_fas)

    # Next, we need to shift all the Row Dimensions below our new rows down by cnt...
    for row in range(len(self.row_dimensions)-1+cnt,row_idx+cnt,-1):
        new_rd = copy.copy(self.row_dimensions[row-cnt])
        new_rd.index = row
        self.row_dimensions[row] = new_rd
        del self.row_dimensions[row-cnt]

    # Now, create our new rows, with all the pretty cells
    row_idx += 1
    for row in range(row_idx,row_idx+cnt):
        # Create a Row Dimension for our new row
        new_rd = copy.copy(self.row_dimensions[row-1])
        new_rd.index = row
        self.row_dimensions[row] = new_rd
        #  for col in range(1,self.max_column):
        for col in range(1,self.max_column+1):
            col = get_column_letter(col)
            #  cell = self.cell('%s%d'%(col,row))
            cell = self['%s%d'%(col,row)]
            cell.value = None
            #  source = self.cell('%s%d'%(col,row-1))
            source = self['%s%d'%(col,row-1)]
            if copy_style:
                cell.number_format = source.number_format
                cell.font      = source.font.copy()
                cell.alignment = source.alignment.copy()
                cell.border    = source.border.copy()
                cell.fill      = source.fill.copy()


    # Check for Merged Cell Ranges that need to be expanded to contain new cells
    for cr_idx, cr in enumerate(self.merged_cell_ranges):
        self.merged_cell_ranges[cr_idx] = CELL_RE.sub(replace,cr)



#  from distutils.core import setup
from setuptools import setup

setup(
    name = 'explay_testr',
    packages = ['explay'],
    scripts = ['bin/exp'],
    version = '0.4.0',
    description = 'make excel jobs playful again',
    author = 'zealzel',
    author_email = 'zealzel@gmail.com',
    url = 'https://github.com/zealzel/explay',
    download_url = 'https://github.com/zealzel/explay/tarball/v0.4.0',
    #  keywords = ['pandas', 'openpyxl', 'fire'],
    #  classifiers = [],
    install_requires=[
        'pandas',
        'openpyxl',
        'regex',
        'fire',
        'XlsxWriter',
        'numpy',
        'xlrd',
        'pyyaml',
    ]
)
